:- module(bibd).

:- export
        bibd/6,
        none_problem/2,
        problem/2,
        problem_symmetries/2,
        gap_sbds_problem/2,
        gap_sbdd_problem/2,
        isbds_problem/2,
        static_problem/2,
        sbdd_ssb_problem/4.

:- lib(ic).
:- lib(ic_global_gac).
:- lib(ic_sbds).
:- lib(ic_gap_sbds).
:- lib(ic_gap_sbdd).

:- lib(ldsb).
:- lib(ldsb_syms).
:- lib(ldsb_sbds).
:- lib(ldsb_static).
:- lib(ldsb_misc).

ldsb_indomain_max(X) :- nonvar(X), !.
ldsb_indomain_max(X) :-
        ic:is_solver_var(X), !,
        get_max(X,V),
        ldsb_try(X, V, _),
        ldsb_indomain(X).


bibd(V, B, R, K, Lambda, M) :-
        %% V * B matrix.
        dim(M, [V, B]),
        M[1..V,1..B] :: 0..1,

        %% Each row should sum to R: there are R occurrences of each
        %% thing.
        ( for(I, 1, V), param(M, R, B) do
            R #= sum(M[I, 1..B]) ),

        %% Each column should sum to K: each block has K things in it.
        ( for(I, 1, B), param(M, K, V) do
            K #= sum(M[1..V, I]) ),

        %% For any distinct rows r1 and r2, the scalar product of r1
        %% and r2 equals Lambda.
        ( for(I, 1, V), param(M, B, Lambda, V) do
          ( for(J, I+1, V), param(M, B, Lambda, I) do
            scalar_product(M[I, 1..B], M[J, 1..B], Lambda) ) ),

        true,
        % % square bibd
        % ( for(I, 1, V), param(M, B, Lambda, V) do
        %   ( for(J, I+1, V), param(M, B, Lambda, I) do
        %     scalar_product(M[1..B, I], M[1..B, J], Lambda) ) ),

        % ( for(I, 2, V), param(M, B) do
        %     collection_to_list(M[I, 1..B], L1),
        %     collection_to_list(M[I-1, 1..B], L2),
        %     lex_le(L1, L2)
        % ),

        % ( for(I, 2, B), param(M, V) do
        %     collection_to_list(M[1..V, I], L1),
        %     collection_to_list(M[1..V, I-1], L2),
        %     lex_le(L1, L2)
        % ),
        true.

scalar_product(Collection1, Collection2, Result) :-
        collection_to_list(Collection1, L1),
        collection_to_list(Collection2, L2),

        length(L1, Length),
        length(Products, Length),
        Products :: 0..1,

        ( foreach(X1, L1),
          foreach(X2, L2),
          foreach(P, Products) do
            P #= X1 * X2 ),

        Result #= sum(Products).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

problem_symmetries(Matrix, Syms) :-
%        rows_interchange(Matrix, RowSym),
%        columns_interchange(Matrix, ColSym),
        RowSym = rows_interchange,
        ColSym = columns_interchange,
        simplify_symmetries(Matrix, [RowSym, ColSym], Syms).

problem([V,B,R,K,Lambda], Matrix) :-
        bibd(V,B,R,K,Lambda,Matrix),
        problem_symmetries(Matrix, Syms),
        ldsb_initialise(Matrix, Syms).

none_problem([V,B,R,K,L], M) :-
        bibd(V,B,R,K,L, M).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gap_sbds_problem([V,B,R,K,L], M) :-
        bibd(V,B,R,K,L, M),
        ic_gap_sbds:sbds_initialise(M, [rows, cols], values:0..1,
                        [symmetry(s_n, [rows], []),
                         symmetry(s_n, [cols], [])],
                        []).

gap_sbdd_problem([V,B,R,K,L], M) :-
        bibd(V,B,R,K,L, M),
        sbdd_initialise(M, [rows, cols], values:0..1,
                        [symmetry(s_n, [rows], []),
                         symmetry(s_n, [cols], [])],
                        []).

isbds_problem([V,B,R,K,L], M) :-
        bibd(V,B,R,K,L, M),
        problem_symmetries(M,Syms),
        convert_symmetries(pairs, M, Syms, SBDSSyms),
        ic_sbds:sbds_initialise(M, 2, SBDSSyms, #=, []).

static_problem([V,B,R,K,Lambda], Matrix) :-
        bibd(V,B,R,K,Lambda,Matrix),
        problem_symmetries(Matrix, Syms),
        post_symmetries(Matrix, Syms).

sbdd_ssb_problem([V,B,R,K,Lambda], Vector, VarPart, ValPart) :-
        bibd(V,B,R,K,Lambda,Matrix),
        problem_symmetries(Matrix, Syms),
        symmetries_to_ssb(Matrix, Syms, Vector, VarPart, ValPart).

