:- module(golf).

:- export
        golf/4,
        none_problem/2,
        problem/2,
        problem_symmetries/2,
        gap_sbds_problem/2,
        gap_sbdd_problem/2,
        isbds_problem/2.

:- lib(ic).
:- lib(ic_sbds).
:- lib(ic_sets).
:- lib(ic_gap_sbds).
:- lib(ic_gap_sbdd).
:- import subset/2 from ic_sets.

:- lib(ldsb).
:- lib(ldsb_syms).
:- lib(ldsb_sbds).

golf(RoundNum,RoundSize,GroupSize,Matrix) :-
	( for(I,1,GroupSize*RoundSize), foreach(I,SetOfAllPlayers) do true ),

	( for(_,1,RoundNum),
	  foreach(GroupsInRound,Rounds),
	  param(SetOfAllPlayers,RoundSize,GroupSize)
	do
	    ( foreach(S,GroupsInRound),
	      for(_,1,RoundSize),
	      param(SetOfAllPlayers,GroupSize)
	    do
		    S subset SetOfAllPlayers,
		    #(S,GroupSize)
	    ),  
	    all_disjoint(GroupsInRound)
	), 
	( fromto(Rounds,[R|Rest0],Rest0,[])
	do
		flatten(Rest0,Rest),
		( foreach(Group,R),
		  param(Rest)
		do
			( foreach(Group1,Rest),
			  param(Group)
			do
				InCommon #:: 0..1,
				#(Group /\ Group1,InCommon)
			)
		)
	),

        dim(Matrix, [RoundNum, RoundSize]),
        ( count(I,1,_), foreach(Round, Rounds), param(Matrix) do
          ( count(J,1,_), foreach(V, Round), param(I, Matrix) do
            subscript(Matrix, [I,J], V)
          )
        ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

problem_symmetries(Matrix, Syms) :-
        % Weeks are interchangeable.
        rows_interchange(Matrix, WeekSym),
        % Golfers are interchangeable.
        set_values_interchange(Matrix, ValSym),
        % Within each week, the groups are interchangeable.
        dim(Matrix, [W,G]),
        ( for(I, 1, W), foreach(Subsym, Subsyms), param(Matrix,G) do
            subscript(Matrix, [I, 1..G], Submatrix),
            variables_interchange(Submatrix, Subsym) ),

        Syms = [WeekSym,ValSym|Subsyms].

problem([W,G,P], Matrix) :-
        golf(W,G,P,Matrix),
        problem_symmetries(Matrix, Syms),
        ldsb_initialise(Matrix, Syms).

none_problem([W,G,P], Matrix) :- golf(W,G,P, Matrix).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gap_sbds_problem([W,G,P], M) :-
        golf(W,G,P, M),
        NumGolfers is G*P,
        ic_gap_sbds:sbds_initialise_set(M, [weeks, groups], value:1..NumGolfers,
                        [symmetry(s_n, [weeks], []),
                         symmetry(s_n, [groups], [weeks:1]),
                         symmetry(s_n, [value], [])],
                        []).

gap_sbdd_problem([W,G,P], M) :-
        golf(W,G,P, M),
        NumGolfers is G*P,
        ic_gap_sbdd:sbdd_initialise_set(M, [weeks, groups], value:1..NumGolfers,
                        [symmetry(s_n, [weeks], []),
                         symmetry(s_n, [groups], [weeks:1]),
                         symmetry(s_n, [value], [])],
                        []).

my_in(S,V,B) :- in(V,S,B).

isbds_problem([W,G,P], M) :-
        golf(W,G,P, M),
        problem_symmetries(M, Syms),
        convert_symmetries(pairs, M, Syms, SBDSSyms),
        ic_sbds:sbds_initialise(M, 2, SBDSSyms, my_in, []).
